#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing pre-releases based on Pipfile configuration in Thoth section."""

import itertools

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import SelectiveCutPreReleasesSieve
from thoth.python import Source
from thoth.python import PackageVersion
from thoth.python import Project

from ..base import AdviserUnitTestCase


class TestSelctiveCutPreReleasesSieve(AdviserUnitTestCase):
    """Test removing pre-releases based on Pipfile configuration in Thoth section."""

    UNIT_TESTED = SelectiveCutPreReleasesSieve

    _CASE_GLOBAL_DISALLOWED_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = false

[thoth.allow_prereleases]
tensorflow = true
flask = false
"""

    _CASE_GLOBAL_DISALLOWED_EMPTY_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = false
"""

    _CASE_GLOBAL_ALLOWED_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = true

[thoth.allow_prereleases]
tensorflow = true
flask = false
"""

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.STABLE,
            RecommendationType.PERFORMANCE,
            RecommendationType.SECURITY,
            RecommendationType.LATEST,
            RecommendationType.TESTING,
        ],
    )
    def test_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
    ) -> None:
        """Test including this pipeline unit."""
        builder_context.recommendation_type = recommendation_type
        builder_context.project = Project.from_strings(self._CASE_GLOBAL_DISALLOWED_PIPFILE)

        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [
            {
                "package_name": None,
                "allow_prereleases": {
                    "tensorflow": True,
                    "flask": False,
                },
            }
        ]

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project = Project.from_strings(self._CASE_GLOBAL_DISALLOWED_PIPFILE)
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.skip(reason="The pipeline unit configuration is specific to Pipfile configuration")
    def test_default_configuration(self, builder_context: PipelineBuilderContext) -> None:
        """Test the default configuration."""

    @pytest.mark.parametrize(
        "recommendation_type,case",
        itertools.product(
            [
                RecommendationType.STABLE,
                RecommendationType.PERFORMANCE,
                RecommendationType.SECURITY,
                RecommendationType.LATEST,
                RecommendationType.TESTING,
            ],
            [_CASE_GLOBAL_ALLOWED_PIPFILE, _CASE_GLOBAL_DISALLOWED_EMPTY_PIPFILE],
        ),
    )
    def test_not_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        case: str,
    ) -> None:
        """Test not including this pipeline unit."""
        builder_context.recommendation_type = recommendation_type
        builder_context.project = Project.from_strings(self._CASE_GLOBAL_ALLOWED_PIPFILE)

        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    @pytest.mark.parametrize("package_name,package_version", [("tensorflow", "2.4.0rc1"), ("flask", "1.0")])
    def test_remove_pre_releases_disallowed_noop(
        self, context: Context, package_name: str, package_version: str
    ) -> None:
        """Test NOT removing dependencies based on pre-release configuration."""
        pv = PackageVersion(
            name=package_name,
            version=f"=={package_version}",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        project = Project.from_strings(self._CASE_GLOBAL_DISALLOWED_PIPFILE)
        context.project = project
        sieve = self.UNIT_TESTED()
        sieve.update_configuration({"package_name": None, "allow_prereleases": project.pipfile.thoth.allow_prereleases})

        assert not context.stack_info
        with self.UNIT_TESTED.assigned_context(context):
            assert list(sieve.run(p for p in [pv])) == [pv]

        assert not context.stack_info, "No stack information should be provided"

    @pytest.mark.parametrize(
        "package_name,package_version",
        [
            ("flask", "1.0dev0"),  # Disabled explicitly.
            ("numpy", "1.0dev0"),  # Disabled implicitly.
        ],
    )
    def test_pre_releases_disallowed_removal(self, context: Context, package_name: str, package_version: str) -> None:
        """Test no removals if pre-releases are allowed."""
        pv = PackageVersion(
            name=package_name,
            version=f"=={package_version}",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        project = Project.from_strings(self._CASE_GLOBAL_DISALLOWED_PIPFILE)
        context.project = project
        sieve = self.UNIT_TESTED()
        sieve.update_configuration({"package_name": None, "allow_prereleases": project.pipfile.thoth.allow_prereleases})

        assert not context.stack_info
        with self.UNIT_TESTED.assigned_context(context):
            assert list(sieve.run(p for p in [pv])) == []

        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing steps."""

import attr
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
import logging

from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required
from thoth.python import PackageVersion

from thoth.adviser.state import State
from packaging.specifiers import SpecifierSet
from .unit import UnitPrescription
from .schema import PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA
from .schema import PRESCRIPTION_STEP_RUN_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class StepPrescription(UnitPrescription):
    """Step base class implementation.

    Configuration option `multi_package_resolution` states whether a step should be run if package
    is resolved multiple times for the same stack.
    """

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("match"): PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA,
            Required("multi_package_resolution"): bool,
            Required("run"): PRESCRIPTION_STEP_RUN_SCHEMA,
        }
    )

    SCORE_MAX = 1.0
    SCORE_MIN = -1.0

    _specifier = attr.ib(type=Optional[SpecifierSet], kw_only=True, init=False, default=None)
    _index_url = attr.ib(type=Optional[str], kw_only=True, init=False, default=None)
    _develop = attr.ib(type=Optional[bool], kw_only=True, init=False, default=None)

    @staticmethod
    def is_step_unit_type() -> bool:
        """Check if this unit is of type step."""
        return True

    @staticmethod
    def _yield_should_include(unit_prescription: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """Yield for every entry stated in the match field."""
        match = unit_prescription["match"]
        run = unit_prescription["run"]
        if isinstance(match, list):
            for item in match:
                yield {
                    "package_name": item["package_version"].get("name"),
                    "multi_package_resolution": run.get("multi_package_resolution", False),
                    "match": item,
                    "run": run,
                }
        else:
            yield {
                "package_name": match["package_version"].get("name"),
                "multi_package_resolution": run.get("multi_package_resolution", False),
                "match": match,
                "run": run,
            }

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include(prescription)
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Prepare before running this pipeline unit."""
        package_version = self.match_prescription.get("package_version", {})
        version_specifier = package_version.get("version")
        if version_specifier:
            self._specifier = SpecifierSet(version_specifier)

        self._index_url = package_version.get("index_url")
        self._prepare_justification_link(self.run_prescription.get("justification", []))
        self._develop = package_version.get("develop")
        super().pre_run()

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Run main entry-point for steps to filter and score packages."""
        if not self._index_url_check(self._index_url, package_version.index.url):
            return None

        if self._specifier and package_version.locked_version not in self._specifier:
            return None

        if self._develop is not None and package_version.develop != self._develop:
            return None

        if not self._run_state_with_initiator(state, package_version):
            return None

        self._run_base()

        score = self.run_prescription.get("score")
        justification = self.run_prescription.get("justification")

        return score, justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing legacy versions from the resolution process."""

from thoth.adviser.context import Context
from thoth.adviser.sieves import LegacyVersionSieve
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserUnitTestCase


class TestLegacyVersionSieve(AdviserUnitTestCase):
    """Test removing legacy versions from the resolution process."""

    UNIT_TESTED = LegacyVersionSieve

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_include(
        self,
        builder_context: PipelineBuilderContext,
    ) -> None:
        """Test including this pipeline unit."""
        builder_context.recommendation_type = RecommendationType.LATEST
        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    def test_remove_legacy_version_noop(self, context: Context) -> None:
        """Test not removing not legacy versions."""
        pv1 = PackageVersion(
            name="flask",
            version="==0.0dev0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        pv2 = PackageVersion(
            name="tensorflow",
            version="==2.0.0rc0",
            index=pv1.index,
            develop=False,
        )

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            assert list(sieve.run(p for p in [pv1, pv2])) == [pv1, pv2]

    def test_remove_legacy_version(self, context: Context) -> None:
        """Test removing legacy versions."""
        pv1 = PackageVersion(
            name="flask",
            version="==foo",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            assert list(sieve.run(p for p in [pv1])) == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test adviser's context passed to pipeline units."""

import flexmock

from .base import AdviserTestCase

from thoth.adviser.dm_report import DependencyMonkeyReport


class TestDependencyMonkeyReport(AdviserTestCase):
    """Test dependency monkey reporting."""

    def test_to_dict(self) -> None:
        """Check dependency monkey report serialization."""
        report = DependencyMonkeyReport()

        assert report.to_dict() == {"skipped": 0, "responses": []}

        product_dict = {"foo": 1}
        product = flexmock()
        product.should_receive("to_dict").with_args().and_return(product_dict).once()

        response = "assaintuosuco"
        report.skipped += 1
        report.add_response(response, product)

        assert report.to_dict() == {
            "skipped": 1,
            "responses": [{"response": response, "product": product_dict}],
        }

    def test_add_responses(self) -> None:
        """Test adding new responses to report."""
        report = DependencyMonkeyReport()

        assert len(report.to_dict()["responses"]) == 0

        response1 = "assaintuosuco"
        product_dict1 = {"foo": 1}
        product1 = flexmock()
        product1.should_receive("to_dict").with_args().and_return(product_dict1).once()

        report.add_response(response1, product1)
        assert len(report.to_dict()["responses"]) == 1
        assert report.to_dict()["responses"][-1]["response"] == response1
        assert report.to_dict()["responses"][-1]["product"] is product_dict1

        response2 = "epoximise"
        product_dict2 = {"foo": 2}
        product2 = flexmock()
        product2.should_receive("to_dict").with_args().and_return(product_dict2).once()

        report.add_response(response2, product2)
        assert len(report.to_dict()["responses"]) == 2
        assert report.to_dict()["responses"][-2]["response"] == response1
        assert report.to_dict()["responses"][-2]["product"] is product_dict1
        assert report.to_dict()["responses"][-1]["response"] == response2
        assert report.to_dict()["responses"][-1]["product"] is product_dict2



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that provides information derived from Python trove classifiers."""

import logging
from enum import Enum
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from ..state import State
from ..wrap import Wrap
from thoth.adviser.enums import RecommendationType
from thoth.common import get_justification_link as jl

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


class _DevelopmentStatusEnum(Enum):
    """Development status trove classifier representation."""

    PLANNING = 1
    PRE_ALPHA = 2
    ALPHA = 3
    BETA = 4
    STABLE = 5
    MATURE = 6
    INACTIVE = 7

    @classmethod
    def from_str(cls, trove_classifier: str) -> Optional["_DevelopmentStatusEnum"]:
        """Get development status enum from a string representation."""
        trove_classifier.strip()

        if trove_classifier == "Development Status:: 1 - Planning":
            return cls.PLANNING

        if trove_classifier == "Development Status:: 2 - Pre - Alpha":
            return cls.PRE_ALPHA

        if trove_classifier == "Development Status:: 3 - Alpha":
            return cls.ALPHA

        if trove_classifier == "Development Status:: 4 - Beta":
            return cls.BETA

        if trove_classifier == "Development Status:: 5 - Production / Stable":
            return cls.STABLE

        if trove_classifier == "Development Status:: 6 - Mature":
            return cls.MATURE

        if trove_classifier == "Development Status:: 7 - Inactive":
            return cls.INACTIVE

        return None


class TroveClassifiersWrap(Wrap):
    """A wrap that provides information derived from Python trove classifiers."""

    _ENVIRONMENT_GPU_CUDA_PREFIX = "Environment :: GPU :: NVIDIA CUDA :: "
    _PROGRAMMING_LANGUAGE_PYTHON_PREFIX = "Programming Language :: Python :: "
    _DEVELOPMENT_STATUS_PREFIX = "Development Status:: "

    _JUSTIFICATION_LINK_PYTHON_VERSION = jl("trove_py")
    _JUSTIFICATION_LINK_USING_UNSTABLE = jl("trove_unstable")
    _JUSTIFICATION_LINK_USING_INACTIVE = jl("trove_inactive")
    _JUSTIFICATION_LINK_CUDA_VERSION = jl("trove_cuda")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[Any, Any], None, None]:
        """Include this wrap in adviser, once."""
        if builder_context.is_adviser_pipeline() and not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def _environment_gpu_cuda_version(self, package_name: str, state: State, trove_classifiers: List[str]) -> None:
        """Add justifications relevant relevant information for GPU CUDA."""
        cuda_version = self.context.project.runtime_environment.cuda_version
        if not cuda_version:
            return None

        cuda_version_trove_classifiers = []
        for trove_classifier in trove_classifiers:
            if not trove_classifier.startswith(self._ENVIRONMENT_GPU_CUDA_PREFIX):
                continue

            cuda_version_trove_classifier = trove_classifier[len(self._ENVIRONMENT_GPU_CUDA_PREFIX) :].strip()

            cuda_version_trove_classifier_parts = cuda_version_trove_classifier.split(".")
            if len(cuda_version_trove_classifier_parts) != 2:
                _LOGGER.debug("Unknown CUDA trove classifier: %r", cuda_version_trove_classifier)
                continue

            cuda_version_trove_classifiers.append(cuda_version_trove_classifier)

        if cuda_version_trove_classifiers and cuda_version not in cuda_version_trove_classifiers:
            state.justification.append(
                {
                    "type": "WARNING",
                    "link": self._JUSTIFICATION_LINK_CUDA_VERSION,
                    "message": f"No CUDA specific trove classifier matching CUDA version used {cuda_version!r}",
                    "package_name": package_name,
                }
            )

    def _development_status(self, package_name: str, state: State, trove_classifiers: List[str]) -> None:
        """Add justifications relevant for development status."""
        for trove_classifier in trove_classifiers:
            development_status = _DevelopmentStatusEnum.from_str(trove_classifier)
            if development_status is None:
                continue

            recommendation_type: RecommendationType = self.context.recommendation_type
            if self.context.recommendation_type not in (RecommendationType.TESTING, RecommendationType.LATEST):
                if development_status is not None and development_status.value < _DevelopmentStatusEnum.STABLE.value:
                    state.justification.append(
                        {
                            "type": "WARNING",
                            "link": self._JUSTIFICATION_LINK_USING_UNSTABLE,
                            "message": f"Development status stated in trove classifiers is "
                            f"{development_status.name.lower().replace('_', '-')!r} which might "
                            f"be not suitable to be used with recommendation type {recommendation_type.name.lower()!r}",
                            "package_name": package_name,
                        }
                    )

            if development_status == _DevelopmentStatusEnum.INACTIVE:
                state.justification.append(
                    {
                        "type": "WARNING",
                        "link": self._JUSTIFICATION_LINK_USING_INACTIVE,
                        "message": f"Inactive development status of {package_name!r} stated in trove classifiers",
                        "package_name": package_name,
                    }
                )

    def _programming_language_python(self, package_name: str, state: State, trove_classifiers: List[str]) -> None:
        """Add justification relevant for Python programming language."""
        python_version = self.context.project.runtime_environment.python_version
        if not python_version:
            return None

        py_trove_classifiers = []
        for trove_classifier in trove_classifiers:
            if not trove_classifier.startswith(self._PROGRAMMING_LANGUAGE_PYTHON_PREFIX):
                continue

            trove_python_version = trove_classifier[len(self._PROGRAMMING_LANGUAGE_PYTHON_PREFIX) :].strip()

            trove_python_version_parts = trove_python_version.split(".")
            if len(trove_python_version_parts) != 2:
                continue

            py_trove_classifiers.append(trove_python_version)

        if py_trove_classifiers and python_version not in py_trove_classifiers:
            state.justification.append(
                {
                    "type": "WARNING",
                    "link": self._JUSTIFICATION_LINK_PYTHON_VERSION,
                    "message": f"No Python specific trove classifier matching Python version used {python_version!r}",
                    "package_name": package_name,
                }
            )

    def run(self, state: State) -> None:
        """Add information derived from Python trove classifiers."""
        runtime_environment = self.context.project.runtime_environment
        operating_system = runtime_environment.operating_system
        os_name = operating_system.name
        os_version = operating_system.version
        python_version = runtime_environment.python_version

        for package_version_tuple in state.resolved_dependencies.values():
            package_name = package_version_tuple[0]

            trove_classifiers = [
                i.strip()
                for i in self.context.graph.get_python_package_version_trove_classifiers_all(
                    package_name=package_name,
                    package_version=package_version_tuple[1],
                    index_url=package_version_tuple[2],
                    os_name=os_name,
                    os_version=os_version,
                    python_version=python_version,
                )
            ]

            if not trove_classifiers:
                continue

            self._environment_gpu_cuda_version(package_name, state, trove_classifiers)
            self._development_status(package_name, state, trove_classifiers)
            self._programming_language_python(package_name, state, trove_classifiers)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Enum types used in thoth-adviser code."""

from enum import auto
from enum import Enum


class _ExtendedEnum(Enum):
    """A custom enum with extended functionality."""

    @classmethod
    def by_name(cls, name: str) -> "Enum":
        """Retrieve enum based on its name."""
        try:
            return cls.__members__[name.upper()]
        except KeyError as exc:
            raise ValueError(
                "Unknown value for type {}, available: {}",
                cls,
                list(cls.__members__.values()),
            ) from exc


class RecommendationType(_ExtendedEnum):
    """Recommendation generation respecting constraints on package-version level observations."""

    STABLE = auto()
    TESTING = auto()
    LATEST = auto()
    PERFORMANCE = auto()
    SECURITY = auto()


class PythonRecommendationOutput(_ExtendedEnum):
    """Recommendation generation respecting constraints on package-version level observations."""

    PIPENV = auto()
    REQUIREMENTS = auto()


class Ecosystem(_ExtendedEnum):
    """Ecosystem known to recommendation engine."""

    PYTHON = auto()


class DecisionType(_ExtendedEnum):
    """Type of decision used in Dependency Monkey for generating stacks."""

    ALL = auto()
    RANDOM = auto()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of hill climbing in the state space."""

from typing import Callable

import flexmock
from hypothesis import given
from hypothesis.strategies import integers

from thoth.adviser.beam import Beam
from thoth.adviser.predictors import HillClimbing
from thoth.adviser.state import State

from ..base import AdviserTestCase


class TestHillClimbing(AdviserTestCase):
    """Tests related to hill climbing in the state space."""

    @given(
        integers(min_value=1, max_value=256),
    )
    def test_run(self, state_factory: Callable[[], State], state_count: int) -> None:
        """Test running the hill climbing method."""
        state = state_factory()
        beam = Beam()
        for _ in range(state_count):
            cloned_state = state.clone()
            cloned_state.iteration = state.iteration + 1
            beam.add_state(cloned_state)

        predictor = HillClimbing()
        context = flexmock(accepted_final_states_count=33, beam=beam)
        with predictor.assigned_context(context):
            next_state, package_tuple = predictor.run()
            assert next_state is not None
            assert next_state is beam.max()
            assert package_tuple[0] in next_state.unresolved_dependencies
            assert package_tuple in next_state.unresolved_dependencies[package_tuple[0]].values()

    def test_pre_run(self) -> None:
        """Test pre-run initialization."""
        context = flexmock(limit=99)

        predictor = HillClimbing()
        assert predictor._history == []
        predictor._history = [(0.99, 33)]

        with predictor.assigned_context(context):
            predictor.pre_run()
            assert predictor._history == [], "Predictor's history not discarded"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test handling of platform information."""

import pytest

from thoth.adviser.boots import PlatformBoot
from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from ..base import AdviserUnitTestCase


class TestPlatformBoot(AdviserUnitTestCase):
    """Test platform boot."""

    UNIT_TESTED = PlatformBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit."""
        builder_context.should_receive("is_included").and_return(False)
        assert list(PlatformBoot.should_include(builder_context)) == [{}]

        builder_context.should_receive("is_included").and_return(True)
        assert list(PlatformBoot.should_include(builder_context)) == []

    def test_adjust(self, context: Context) -> None:
        """Test adjusting the runtime environment to a default if not provided explicitly."""
        default_platform = "linux-x86_64"
        context.project.runtime_environment.platform = None
        context.graph.should_receive("python_package_version_depends_on_platform_exists").with_args(
            default_platform
        ).and_return(True).once()

        unit = PlatformBoot()
        with unit.assigned_context(context):
            assert unit.run() is None

        assert context.project.runtime_environment.platform == default_platform

    def test_no_platform(self, context: Context) -> None:
        """Test raising an exception if the given platform was not found in the database."""
        platform = "foo-bar"
        context.project.runtime_environment.platform = platform
        context.graph.should_receive("python_package_version_depends_on_platform_exists").with_args(
            platform
        ).and_return(False).once()

        unit = PlatformBoot()
        with unit.assigned_context(context):
            with pytest.raises(NotAcceptable, match="No platform conforming to 'foo-bar' found in the database"):
                unit.run()

    def test_no_adjust(self, context: Context) -> None:
        """Test no adjustment if the runtime environment provides platform info."""
        platform = "linux-i586"
        context.project.runtime_environment.platform = platform
        context.graph.should_receive("python_package_version_depends_on_platform_exists").with_args(
            platform
        ).and_return(True).once()

        unit = PlatformBoot()
        with unit.assigned_context(context):
            assert unit.run() is None

        assert context.project.runtime_environment.platform == platform



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test manipulation with a pipeline product."""


import os
import json
import flexmock
from itertools import chain

from thoth.adviser.product import Product
from thoth.adviser.state import State
from thoth.adviser.context import Context
from thoth.common import get_justification_link as jl
from thoth.common import RuntimeEnvironment
from thoth.python import PackageVersion
from thoth.python import Pipfile
from thoth.python import PipfileLock
from thoth.python import Source

from .base import AdviserTestCase


class TestProduct(AdviserTestCase):
    """Test manipulation with product."""

    _PIPFILE = """
[[source]]
name = "pypi-org"
url = "https://pypi.org/simple"
verify_ssl = true

[dev-packages]

[packages]
tensorflow = "*"

[requires]
python_version = "3.7"

[thoth.allow_prereleases]
black = true
"""

    _DEPENDENCIES_NO_CYCLE = {
        "absl-py": {
            ("absl-py", "0.8.1", "https://pypi.org/simple"): {
                ("six", "1.13.0", "https://pypi.org/simple"),
                ("six", "1.9.0", "https://pypi.org/simple"),
            }
        },
        "astor": {
            ("astor", "0.8.0", "https://pypi.org/simple"): {
                ("six", "1.13.0", "https://pypi.org/simple"),
            }
        },
        "tensorflow": {
            ("tensorflow", "1.9.0", "https://pypi.org/simple"): {
                ("astor", "0.8.0", "https://pypi.org/simple"),
                ("absl-py", "0.8.1", "https://pypi.org/simple"),
            }
        },
    }

    _DEPENDENCIES_CYCLE = {
        "a": {
            ("a", "1.0.0", "https://pypi.org/simple"): {
                ("b", "1.0.0", "https://pypi.org/simple"),
            }
        },
        "b": {
            ("b", "1.0.0", "https://pypi.org/simple"): {
                ("c", "1.1.0", "https://pypi.org/simple"),
            }
        },
        "c": {
            ("c", "1.0.0", "https://pypi.org/simple"): {
                ("a", "1.0.0", "https://pypi.org/simple"),
                ("b", "1.0.0", "https://pypi.org/simple"),
            }
        },
    }

    _DEPENDENCIES_NO_CYCLE_PIPFILE_LOCK = PipfileLock.from_dict(
        {
            "_meta": {
                "hash": {"sha256": "001e85311f65a97fe23d21061c0b68139015dfea5b4bfb8b91551ede367fe1d1"},
                "pipfile-spec": 6,
                "requires": {"python_version": "3.6"},
                "sources": [{"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True}],
            },
            "default": {
                "absl-py": {
                    "hashes": ["sha256:d9129186431e150d7fe455f1cb1ecbb92bb5dba9da9bc3ef7b012d98c4db2526"],
                    "index": "pypi",
                    "version": "==0.8.1",
                },
                "astor": {
                    "hashes": [
                        "sha256:37a6eed8b371f1228db08234ed7f6cfdc7817a3ed3824797e20cbb11dc2a7862",
                    ],
                    "index": "pypi",
                    "version": "==0.8.0",
                },
                "six": {
                    "hashes": [
                        "sha256:13f9f196f330c7c2c5d7a5cf91af894110ca0215ac051b5844701f2bfd934d52",
                    ],
                    "index": "pypi",
                    "version": "==1.13.0",
                },
                "tensorflow": {
                    "hashes": [
                        "sha256:d351f7db08b8de322536c5886fada3e37feae809bfd37368050f9eeea544b87e",
                    ],
                    "index": "pypi",
                    "version": "==1.9.0",
                },
            },
            "develop": {},
        },
        pipfile=None,
    )

    _DEPENDENCIES_CYCLE_PIPFILE_LOCK = PipfileLock.from_dict(
        {
            "_meta": {
                "hash": {"sha256": "FOO"},
                "pipfile-spec": 6,
                "requires": {"python_version": "3.8"},
                "sources": [{"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True}],
            },
            "default": {
                "a": {
                    "hashes": ["sha256:foo"],
                    "index": "pypi",
                    "version": "==1.0.0",
                },
                "b": {
                    "hashes": ["sha256:foo"],
                    "index": "pypi",
                    "version": "==1.0.0",
                },
                "c": {
                    "hashes": ["sha256:foo"],
                    "index": "pypi",
                    "version": "==1.0.0",
                },
            },
            "develop": {},
        },
        pipfile=None,
    )

    def test_from_final_state(self, context: Context) -> None:
        """Test instantiating product from a final state."""
        state = State(
            score=0.5,
            resolved_dependencies={
                "daiquiri": ("daiquiri", "1.6.0", "https://pypi.org/simple"),
                "numpy": ("numpy", "1.17.4", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
        )
        state.add_justification(self.JUSTIFICATION_SAMPLE_1)

        pypi = Source("https://pypi.org/simple")

        pv_daiquiri_locked = PackageVersion(name="daiquiri", version="==1.6.0", index=pypi, develop=False)
        pv_numpy_locked = PackageVersion(name="numpy", version="==1.17.4", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("daiquiri", "1.6.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_daiquiri_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "daiquiri", "1.6.0", "https://pypi.org/simple"
        ).and_return(["000"]).ordered()

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.17.4", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "numpy", "1.17.4", "https://pypi.org/simple"
        ).and_return(["111"]).ordered()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "tensorflow", "2.0.0", "https://pypi.org/simple"
        ).and_return(["222"]).ordered()

        pv_daiquiri = PackageVersion(name="daiquiri", version="*", index=pypi, develop=False)
        pv_tensorflow = PackageVersion(name="tensorflow", version=">=2.0.0", index=pypi, develop=False)

        project = flexmock(
            pipfile=Pipfile.from_string(self._PIPFILE),
            runtime_environment=RuntimeEnvironment.from_dict({"operating_system": {"name": "rhel"}}),
        )
        project.should_receive("iter_dependencies").with_args(with_devel=True).and_return(
            [pv_daiquiri, pv_tensorflow]
        ).once()

        context.project = project
        context.dependencies = {
            "daiquiri": {
                ("daiquiri", "1.6.0", "https://pypi.org/simple"): set(),
            },
            "numpy": {("numpy", "1.17.4", "https://pypi.org/simple"): set()},
            "tensorflow": {
                ("tensorflow", "2.0.0", "https://pypi.org/simple"): {("numpy", "1.17.4", "https://pypi.org/simple")}
            },
        }
        context.dependents = {
            "daiquiri": {
                ("daiquiri", "1.6.0", "https://pypi.org/simple"): set(),
            },
            "numpy": {
                ("numpy", "1.17.4", "https://pypi.org/simple"): {
                    (
                        ("tensorflow", "2.0.0", "https://pypi.org/simple"),
                        "fedora",
                        "31",
                        "3.7",
                    )
                }
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
        }
        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.17.4",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.7'").once()

        assert "THOTH_ADVISER_METADATA" not in os.environ
        metadata_justification = {"thoth.adviser": {"justification": [{"bar": "baz"}]}}
        os.environ["THOTH_ADVISER_METADATA"] = json.dumps(metadata_justification)
        try:
            product = Product.from_final_state(state=state, context=context)
        finally:
            os.environ.pop("THOTH_ADVISER_METADATA")

        assert product.score == state.score
        assert product.justification == list(
            chain(metadata_justification["thoth.adviser"]["justification"], state.justification)
        )
        assert product.advised_runtime_environment == state.advised_runtime_environment
        assert product.project.to_dict() == {
            "constraints": [],
            "requirements": {
                "packages": {
                    "daiquiri": {"index": "pypi-org-simple", "version": "*"},
                    "tensorflow": {"index": "pypi-org-simple", "version": ">=2.0.0"},
                },
                "dev-packages": {},
                "requires": {"python_version": "3.7"},
                "source": [
                    {
                        "url": "https://pypi.org/simple",
                        "verify_ssl": True,
                        "name": "pypi-org",
                    },
                    {
                        "url": "https://pypi.org/simple",
                        "verify_ssl": True,
                        "name": "pypi-org-simple",
                    },
                ],
                "thoth": {
                    "allow_prereleases": {"black": True},
                },
            },
            "requirements_locked": {
                "_meta": {
                    "sources": [
                        {"name": "pypi-org", "url": "https://pypi.org/simple", "verify_ssl": True},
                        {
                            "url": "https://pypi.org/simple",
                            "verify_ssl": True,
                            "name": "pypi-org-simple",
                        },
                    ],
                    "requires": {"python_version": "3.7"},
                    "hash": {"sha256": "6cc8365e799b949fb6cc564cea2d8e0e8a782ab676a006e65abbe14621b93381"},
                    "pipfile-spec": 6,
                },
                "default": {
                    "daiquiri": {
                        "version": "==1.6.0",
                        "hashes": ["sha256:000"],
                        "index": "pypi-org-simple",
                    },
                    "numpy": {
                        "version": "==1.17.4",
                        "hashes": ["sha256:111"],
                        "index": "pypi-org-simple",
                        "markers": "python_version >= '3.7'",
                    },
                    "tensorflow": {
                        "version": "==2.0.0",
                        "hashes": ["sha256:222"],
                        "index": "pypi-org-simple",
                    },
                },
                "develop": {},
            },
            "runtime_environment": {
                "hardware": {"cpu_family": None, "cpu_model": None, "gpu_model": None},
                "operating_system": {"name": "rhel", "version": None},
                "python_version": None,
                "cuda_version": None,
                "cudnn_version": None,
                "name": None,
                "platform": None,
                "base_image": None,
                "mkl_version": None,
                "openblas_version": None,
                "openmpi_version": None,
                "recommendation_type": None,
            },
        }

    def test_to_dict(self, context: Context) -> None:
        """Test conversion of this product into a dictionary representation."""
        project = context.project
        project.should_receive("to_dict").with_args(keep_thoth_section=True).and_return({"baz": "bar"}).once()

        advised_runtime_environment = flexmock()
        advised_runtime_environment.should_receive("to_dict").with_args().and_return({"hello": "thoth"}).once()

        advised_manifest_changes = [
            [
                {
                    "apiVersion": "apps.openshift.io/v1",
                    "kind": "DeploymentConfig",
                    "patch": {
                        "op": "add",
                        "path": "spec.template.spec.containers[0].env",
                        "value": {"name": "OMP_NUM_THREADS", "value": "1"},
                    },
                }
            ]
        ]

        product = Product(
            advised_manifest_changes=advised_manifest_changes,
            advised_runtime_environment=advised_runtime_environment,
            justification=[{"foo": "bar"}],
            project=project,
            score=0.999,
            context=context,
        )

        assert product.to_dict() == {
            "score": 0.999,
            "project": {"baz": "bar"},
            "justification": [{"foo": "bar"}],
            "advised_runtime_environment": {"hello": "thoth"},
            "advised_manifest_changes": advised_manifest_changes,
            "dependency_graph": {
                "edges": [],
                "nodes": [
                    "absl-py",
                    "astor",
                    "click",
                    "flask",
                    "gast",
                    "grpcio",
                    "itsdangerous",
                    "jinja2",
                    "markdown",
                    "markupsafe",
                    "numpy",
                    "protobuf",
                    "six",
                    "tensorboard",
                    "tensorflow",
                    "termcolor",
                    "werkzeug",
                    "wheel",
                ],
            },
        }

    def test_environment_markers(self, context: Context) -> None:
        """Test handling of environment markers across multiple runs."""
        state = State(
            score=0.0,
            resolved_dependencies={
                "numpy": ("numpy", "1.0.0", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
            justification=[{"type": "INFO", "message": "Foo bar", "link": "https://thoth-station.ninja"}],
        )

        # Make sure tested packages are not direct dependencies. In such cases the behaviour is
        # different, see other tests.
        context.project.pipfile.packages.packages.pop("tensorflow", None)
        context.project.pipfile.packages.packages.pop("numpy", None)

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "numpy", "1.0.0", "https://pypi.org/simple"
        ).and_return(["000"]).once()

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "tensorflow", "2.0.0", "https://pypi.org/simple"
        ).and_return(["111"]).once()

        pypi = Source("https://pypi.org/simple")
        pv_numpy_locked = PackageVersion(name="numpy", version="==1.0.0", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).twice()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).twice()

        context.dependents = {
            "numpy": {
                ("numpy", "1.0.0", "https://pypi.org/simple"): {
                    (
                        ("tensorflow", "2.0.0", "https://pypi.org/simple"),
                        "fedora",
                        "31",
                        "3.7",
                    )
                }
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
        }

        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.0.0",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.7'").and_return("python_version >= '3' or 1").twice()

        product = Product.from_final_state(context=context, state=state)
        expected = {
            "advised_manifest_changes": [],
            "advised_runtime_environment": None,
            "justification": [
                {
                    "link": "https://thoth-station.ninja",
                    "message": "Foo bar",
                    "type": "INFO",
                }
            ],
            "dependency_graph": {"edges": [], "nodes": ["numpy", "tensorflow"]},
            "project": {
                "constraints": [],
                "requirements": {
                    "dev-packages": {},
                    "packages": {"flask": "*"},
                    "requires": {"python_version": "3.6"},
                    "source": [
                        {
                            "name": "pypi",
                            "url": "https://pypi.org/simple",
                            "verify_ssl": True,
                        },
                        {
                            "name": "pypi-org-simple",
                            "url": "https://pypi.org/simple",
                            "verify_ssl": True,
                        },
                    ],
                    "thoth": {
                        "allow_prereleases": {},
                        "disable_index_adjustment": False,
                    },
                },
                "requirements_locked": {
                    "_meta": {
                        "hash": {"sha256": "2e49395dfa87159358e581bd22e656c27c0dab04894d1b137a14f85bb387ea51"},
                        "pipfile-spec": 6,
                        "requires": {"python_version": "3.6"},
                        "sources": [
                            {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                            {
                                "name": "pypi-org-simple",
                                "url": "https://pypi.org/simple",
                                "verify_ssl": True,
                            },
                        ],
                    },
                    "default": {
                        "numpy": {
                            "hashes": ["sha256:000"],
                            "index": "pypi-org-simple",
                            "markers": "python_version >= '3.7'",
                            "version": "==1.0.0",
                        },
                        "tensorflow": {
                            "hashes": ["sha256:111"],
                            "index": "pypi-org-simple",
                            "version": "==2.0.0",
                        },
                    },
                    "develop": {},
                },
                "runtime_environment": {
                    "base_image": None,
                    "cuda_version": None,
                    "cudnn_version": None,
                    "hardware": {"cpu_family": None, "cpu_model": None, "gpu_model": None},
                    "name": None,
                    "operating_system": {"name": None, "version": None},
                    "openblas_version": None,
                    "openmpi_version": None,
                    "mkl_version": None,
                    "python_version": None,
                    "recommendation_type": None,
                    "platform": None,
                },
            },
            "score": 0.0,
        }

        assert product.to_dict() == expected

        # Markers should not intersect.
        product = Product.from_final_state(context=context, state=state)
        expected["project"]["requirements_locked"]["default"]["numpy"]["markers"] = "python_version >= '3' or 1"
        assert product.to_dict() == expected

    def test_environment_markers_direct_dependency(self, context: Context) -> None:
        """Test handling environment markers for direct dependencies."""
        state = State(
            score=0.0,
            resolved_dependencies={
                "numpy": ("numpy", "1.0.0", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
            justification=[{"type": "INFO", "message": "Foo bar", "link": "https://thoth-station.ninja"}],
        )

        pypi = Source("https://pypi.org/simple")

        # Let's assume tensorflow is our direct dependency with an environment marker set. It sets also an environment
        # marker for numpy.
        context.project.pipfile.packages.packages.pop("tensorflow", None)
        tf_package_version_direct = PackageVersion(
            name="tensorflow", version=">1.0.0", index=pypi, develop=False, markers="python_version >= '3.6'"
        )
        context.project.pipfile.add_package_version(tf_package_version_direct)
        # Just to make sure numpy is not in the direct dependency listing.
        context.project.pipfile.packages.packages.pop("numpy", None)

        pv_numpy_locked = PackageVersion(name="numpy", version="==1.0.0", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            *pv_numpy_locked.to_tuple()
        ).and_return(["000"]).once()

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            *pv_tensorflow_locked.to_tuple()
        ).and_return(["111"]).once()

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).once()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).once()

        context.dependents = {
            "numpy": {
                ("numpy", "1.0.0", "https://pypi.org/simple"): {
                    (
                        ("tensorflow", "2.0.0", "https://pypi.org/simple"),
                        "fedora",
                        "31",
                        "3.7",
                    )
                }
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
        }

        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.0.0",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.7'").once()

        product = Product.from_final_state(context=context, state=state)
        expected = {
            "advised_manifest_changes": [],
            "advised_runtime_environment": None,
            "justification": [
                {
                    "link": "https://thoth-station.ninja",
                    "message": "Foo bar",
                    "type": "INFO",
                }
            ],
            "dependency_graph": {"edges": [], "nodes": ["numpy", "tensorflow"]},
            "project": {
                "constraints": [],
                "requirements": {
                    "dev-packages": {},
                    "packages": {
                        "flask": "*",
                        "tensorflow": {
                            "index": "pypi-org-simple",
                            "markers": "python_version >= '3.6'",
                            "version": ">1.0.0",
                        },
                    },
                    "requires": {"python_version": "3.6"},
                    "source": [
                        {
                            "name": "pypi",
                            "url": "https://pypi.org/simple",
                            "verify_ssl": True,
                        },
                        {
                            "name": "pypi-org-simple",
                            "url": "https://pypi.org/simple",
                            "verify_ssl": True,
                        },
                    ],
                    "thoth": {
                        "allow_prereleases": {},
                        "disable_index_adjustment": False,
                    },
                },
                "requirements_locked": {
                    "_meta": {
                        "hash": {"sha256": "382cd84046246d08cf670a07c61628958dec76e05d2473bf936866fc54619f9a"},
                        "pipfile-spec": 6,
                        "requires": {"python_version": "3.6"},
                        "sources": [
                            {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                            {
                                "name": "pypi-org-simple",
                                "url": "https://pypi.org/simple",
                                "verify_ssl": True,
                            },
                        ],
                    },
                    "default": {
                        "numpy": {
                            "hashes": ["sha256:000"],
                            "index": "pypi-org-simple",
                            "markers": "python_version >= '3.7'",
                            "version": "==1.0.0",
                        },
                        "tensorflow": {
                            "hashes": ["sha256:111"],
                            "index": "pypi-org-simple",
                            "markers": "python_version >= '3.6'",
                            "version": "==2.0.0",
                        },
                    },
                    "develop": {},
                },
                "runtime_environment": {
                    "base_image": None,
                    "cuda_version": None,
                    "cudnn_version": None,
                    "hardware": {"cpu_family": None, "cpu_model": None, "gpu_model": None},
                    "name": None,
                    "operating_system": {"name": None, "version": None},
                    "openblas_version": None,
                    "openmpi_version": None,
                    "mkl_version": None,
                    "python_version": None,
                    "recommendation_type": None,
                    "platform": None,
                },
            },
            "score": 0.0,
        }

        assert product.to_dict() == expected

    def test_environment_markers_shared(self, context: Context) -> None:
        """Test handling of environment markers when multiple dependencies share one."""
        state = State(
            score=0.0,
            resolved_dependencies={
                "pandas": ("pandas", "1.0.0", "https://pypi.org/simple"),
                "numpy": ("numpy", "1.0.0", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
            justification=[{"type": "INFO", "message": "Foo bar", "link": "https://thoth-station.ninja"}],
        )

        # Make sure tested packages are not direct dependencies. In such cases the behaviour is
        # different, see other tests.
        context.project.pipfile.packages.packages.pop("pandas", None)
        context.project.pipfile.packages.packages.pop("numpy", None)
        context.project.pipfile.packages.packages.pop("tensorflow", None)

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "numpy", "1.0.0", "https://pypi.org/simple"
        ).and_return(["000"]).once()

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "tensorflow", "2.0.0", "https://pypi.org/simple"
        ).and_return(["111"]).once()

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "pandas", "1.0.0", "https://pypi.org/simple"
        ).and_return(["222"]).once()

        pypi = Source("https://pypi.org/simple")
        pv_numpy_locked = PackageVersion(name="numpy", version="==1.0.0", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)
        pv_pandas_locked = PackageVersion(name="pandas", version="==1.0.0", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).once()

        context.should_receive("get_package_version").with_args(
            ("pandas", "1.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_pandas_locked).once()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).once()

        context.dependents = {
            "numpy": {
                ("numpy", "1.0.0", "https://pypi.org/simple"): [  # set to list for reproducible runs.
                    (
                        ("tensorflow", "2.0.0", "https://pypi.org/simple"),
                        "fedora",
                        "31",
                        "3.7",
                    ),
                    (
                        ("pandas", "1.0.0", "https://pypi.org/simple"),
                        "fedora",
                        "31",
                        "3.7",
                    ),
                ]
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
            "pandas": {("pandas", "1.0.0", "https://pypi.org/simple"): set()},
        }

        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.0.0",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.8'").once()

        context.graph.should_receive("get_python_environment_marker").with_args(
            "pandas",
            "1.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.0.0",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return(None).once()

        product = Product.from_final_state(context=context, state=state)
        expected = {
            "advised_manifest_changes": [],
            "advised_runtime_environment": None,
            "dependency_graph": {"edges": [], "nodes": ["pandas", "numpy", "tensorflow"]},
            "justification": [
                {
                    "link": "https://thoth-station.ninja",
                    "message": "Foo bar",
                    "type": "INFO",
                }
            ],
            "project": {
                "constraints": [],
                "requirements": {
                    "dev-packages": {},
                    "packages": {"flask": "*"},
                    "requires": {"python_version": "3.6"},
                    "source": [
                        {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                        {"name": "pypi-org-simple", "url": "https://pypi.org/simple", "verify_ssl": True},
                    ],
                    "thoth": {
                        "allow_prereleases": {},
                        "disable_index_adjustment": False,
                    },
                },
                "requirements_locked": {
                    "_meta": {
                        "hash": {"sha256": "2e49395dfa87159358e581bd22e656c27c0dab04894d1b137a14f85bb387ea51"},
                        "pipfile-spec": 6,
                        "requires": {"python_version": "3.6"},
                        "sources": [
                            {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                            {"name": "pypi-org-simple", "url": "https://pypi.org/simple", "verify_ssl": True},
                        ],
                    },
                    "default": {
                        "numpy": {"hashes": ["sha256:000"], "index": "pypi-org-simple", "version": "==1.0.0"},
                        "pandas": {"hashes": ["sha256:222"], "index": "pypi-org-simple", "version": "==1.0.0"},
                        "tensorflow": {"hashes": ["sha256:111"], "index": "pypi-org-simple", "version": "==2.0.0"},
                    },
                    "develop": {},
                },
                "runtime_environment": {
                    "base_image": None,
                    "cuda_version": None,
                    "cudnn_version": None,
                    "hardware": {"cpu_family": None, "cpu_model": None, "gpu_model": None},
                    "name": None,
                    "operating_system": {"name": None, "version": None},
                    "platform": None,
                    "mkl_version": None,
                    "openmpi_version": None,
                    "openblas_version": None,
                    "python_version": None,
                    "recommendation_type": None,
                },
            },
            "score": 0.0,
        }

        assert product.to_dict() == expected

    def test_no_observation(self, context: Context) -> None:
        """Test adding information about no justification added."""
        state = State(
            score=0.0,
            resolved_dependencies={
                "flask": ("flask", "0.12", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
        )

        context.project.pipfile.packages.packages.pop("tensorflow")

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "flask", "0.12", "https://pypi.org/simple"
        ).and_return(["222"]).once()

        pypi = Source("https://pypi.org/simple")
        pv_pandas_locked = PackageVersion(name="flask", version="==0.12", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("flask", "0.12", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_pandas_locked).once()

        context.dependents = {
            "flask": {("flask", "0.12", "https://pypi.org/simple"): set()},
        }

        product = Product.from_final_state(context=context, state=state)
        expected = {
            "advised_manifest_changes": [],
            "advised_runtime_environment": None,
            "dependency_graph": {"edges": [], "nodes": ["flask"]},
            "justification": [
                {
                    "type": "INFO",
                    "message": "No issues spotted for this stack based on Thoth's database",
                    "link": jl("no_observations"),
                }
            ],
            "project": {
                "constraints": [],
                "requirements": {
                    "dev-packages": {},
                    "packages": {"flask": "*"},
                    "requires": {"python_version": "3.6"},
                    "source": [
                        {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                        {"name": "pypi-org-simple", "url": "https://pypi.org/simple", "verify_ssl": True},
                    ],
                    "thoth": {
                        "allow_prereleases": {},
                        "disable_index_adjustment": False,
                    },
                },
                "requirements_locked": {
                    "_meta": {
                        "hash": {"sha256": "2e49395dfa87159358e581bd22e656c27c0dab04894d1b137a14f85bb387ea51"},
                        "pipfile-spec": 6,
                        "requires": {"python_version": "3.6"},
                        "sources": [
                            {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True},
                            {"name": "pypi-org-simple", "url": "https://pypi.org/simple", "verify_ssl": True},
                        ],
                    },
                    "default": {
                        "flask": {"hashes": ["sha256:222"], "index": "pypi-org-simple", "version": "==0.12"},
                    },
                    "develop": {},
                },
                "runtime_environment": {
                    "base_image": None,
                    "cuda_version": None,
                    "cudnn_version": None,
                    "hardware": {"cpu_family": None, "cpu_model": None, "gpu_model": None},
                    "name": None,
                    "operating_system": {"name": None, "version": None},
                    "platform": None,
                    "mkl_version": None,
                    "openmpi_version": None,
                    "openblas_version": None,
                    "python_version": None,
                    "recommendation_type": None,
                },
            },
            "score": 0.0,
        }

        assert product.to_dict() == expected

    def test_construct_dependency_graph_basic(self) -> None:
        """Test constructing dependency graph."""
        context = flexmock(dependencies=self._DEPENDENCIES_NO_CYCLE)
        dependency_graph = Product._construct_dependency_graph(context, self._DEPENDENCIES_NO_CYCLE_PIPFILE_LOCK)
        assert dependency_graph["nodes"] == ["absl-py", "astor", "six", "tensorflow"]
        assert set(tuple(i) for i in dependency_graph["edges"]) == {(0, 2), (1, 2), (3, 1), (3, 0)}

    def test_construct_dependency_graph_cycle(self, context: Context) -> None:
        """Test constructing dependency graph information with cycles."""
        context = flexmock(dependencies=self._DEPENDENCIES_CYCLE)
        dependency_graph = Product._construct_dependency_graph(context, self._DEPENDENCIES_CYCLE_PIPFILE_LOCK)
        assert dependency_graph["nodes"] == ["a", "b", "c"]
        assert set(tuple(i) for i in dependency_graph["edges"]) == {(0, 1), (2, 0), (2, 1)}



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A generic unit representatives used for testing.."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

from thoth.adviser.wrap import Wrap
from thoth.adviser.state import State
from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required

if TYPE_CHECKING:
    from thoth.adviser.pipeline_builder import PipelineBuilderContext


class Wrap1(Wrap):
    """A testing wrap implementation."""

    CONFIGURATION_DEFAULT = {
        "thoth": [2018, 2019],
        "cities": ["Brno", "Bonn", "Boston", "Milan"],
        "package_name": None,
    }
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("thoth"): object, Required("cities"): object, Required("package_name"): SchemaAny(str)}
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self, state: State) -> None:
        """Run noop method."""


class Wrap2(Wrap):
    """A testing wrap implementation."""

    CONFIGURATION_DEFAULT: Dict[str, Any] = {"package_name": "hexsticker"}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str})

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self, state: State) -> None:
        """Run noop method."""


__all__ = ["Wrap1", "Wrap2"]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests related to check if GPU specific recommendations should be done."""

from thoth.adviser.boots import GPUBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext

from ..base import AdviserUnitTestCase


class TestGPUBoot(AdviserUnitTestCase):
    """Test a boot that checks if GPU specific recommendations should be done."""

    UNIT_TESTED = GPUBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.project.runtime_environment.hardware.gpu_model = 0x42
        builder_context.project.runtime_environment.cuda_version = None
        self.verify_multiple_should_include(builder_context)

    def test_include_no_cuda(self, builder_context: PipelineBuilderContext) -> None:
        """Verify including when no CUDA is supplied."""
        builder_context.project.runtime_environment.hardware.gpu_model = 0x42
        builder_context.project.runtime_environment.cuda_version = None
        self.verify_multiple_should_include(builder_context)

    def test_include_no_gpu(self, builder_context: PipelineBuilderContext) -> None:
        """Verify including when no GPU is supplied."""
        builder_context.project.runtime_environment.hardware.gpu_model = None
        builder_context.project.runtime_environment.cuda_version = "10.0"
        self.verify_multiple_should_include(builder_context)

    def test_no_cuda(self, context: Context) -> None:
        """Test when no CUDA is supplied, but GPU is available."""
        context.project.runtime_environment.hardware.gpu_model = 0x42
        context.project.runtime_environment.cuda_version = None

        boot = self.UNIT_TESTED()
        with boot.assigned_context(context):
            boot.run()

        assert context.stack_info, "No stack info provided"
        assert context.project.runtime_environment.hardware.gpu_model is None
        assert context.project.runtime_environment.cuda_version is None
        assert self.verify_justification_schema(context.stack_info) is True

    def test_no_gpu(self, context: Context) -> None:
        """Test when no GPU is available, but CUDA is available."""
        context.project.runtime_environment.hardware.gpu_model = None
        context.project.runtime_environment.cuda_version = "9.0"

        boot = self.UNIT_TESTED()
        with boot.assigned_context(context):
            boot.run()

        assert context.stack_info, "No stack info provided"
        assert context.project.runtime_environment.hardware.gpu_model is None
        assert context.project.runtime_environment.cuda_version is None
        assert self.verify_justification_schema(context.stack_info) is True



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing sieves."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import TYPE_CHECKING
import logging

import attr
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required

from packaging.specifiers import SpecifierSet
from .unit import UnitPrescription
from .schema import PRESCRIPTION_SIEVE_MATCH_ENTRY_SCHEMA
from .schema import PRESCRIPTION_SIEVE_RUN_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SievePrescription(UnitPrescription):
    """Sieve base class implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("match"): PRESCRIPTION_SIEVE_MATCH_ENTRY_SCHEMA,
            Required("run"): SchemaAny(PRESCRIPTION_SIEVE_RUN_SCHEMA, None),
        }
    )

    _logged = attr.ib(type=bool, kw_only=True, init=False, default=False)
    _specifier = attr.ib(type=Optional[SpecifierSet], kw_only=True, init=False, default=None)
    _index_url = attr.ib(type=Optional[str], kw_only=True, init=False, default=None)
    _develop = attr.ib(type=Optional[str], kw_only=True, init=False, default=None)

    @staticmethod
    def is_sieve_unit_type() -> bool:
        """Check if this unit is of type sieve."""
        return True

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include(prescription)
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Prepare before running this pipeline unit."""
        package_version = self.match_prescription.get("package_version", {})
        version_specifier = package_version.get("version")
        if version_specifier:
            self._specifier = SpecifierSet(version_specifier)
            self._specifier.prereleases = True

        self._index_url = package_version.get("index_url")
        self._logged = False
        self._develop = package_version.get("develop")
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Run main entry-point for sieves to filter and score packages."""
        for package_version in package_versions:
            if (
                (not self._specifier or package_version.locked_version in self._specifier)
                and self._index_url_check(self._index_url, package_version.index.url)
                and (self._develop is None or self._develop == package_version.develop)
            ):
                if not self._logged:
                    self._logged = True
                    self._run_log()
                    self._run_stack_info()

                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing adviser's test cases."""

from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
import os

from voluptuous import All
from voluptuous import Any as SchemaAny
from voluptuous import Invalid
from voluptuous import Length
from voluptuous import Optional as SchemaOptional
from voluptuous import Schema
import pytest

from thoth.adviser.boot import Boot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.pseudonym import Pseudonym
from thoth.adviser.sieve import Sieve
from thoth.adviser.step import Step
from thoth.adviser.stride import Stride
from thoth.adviser.unit import Unit
from thoth.adviser.wrap import Wrap


class AdviserTestCaseException(Exception):  # noqa: N818
    """A base class for exceptions that can occur in the test suite."""


class AdviserJustificationSchemaError(AdviserTestCaseException):
    """An exception raised when the justification reported violates schema."""


class AdviserTestCase:
    """A base class for implementing adviser's test cases."""

    data_dir = Path(os.path.dirname(os.path.realpath(__file__))) / "data"

    JUSTIFICATION_SAMPLE_1 = [
        {"message": "Justification sample 1", "type": "WARNING", "link": "https://thoth-station.ninja"},
        {"message": "Justification sample 1", "type": "INFO", "link": "https://thoth-station.ninja"},
        {"message": "Justification sample 1", "type": "ERROR", "link": "https://thoth-station.ninja"},
    ]

    JUSTIFICATION_SAMPLE_2 = [
        {
            "message": "Justification sample 2",
            "type": "INFO",
            "link": "https://thoth-station.ninja",
            "advisory": "Bark!",
        },
    ]

    JUSTIFICATION_SAMPLE_3 = [
        {
            "message": "Justification sample 2",
            "type": "INFO",
            "link": "https://thoth-station.ninja",
            "package_name": "tensorflow",
            "version_range": "<2.3>=",
        },
    ]

    _JUSTIFICATION_SCHEMA = Schema(
        [
            {
                "message": All(str, Length(min=1)),
                "type": SchemaAny("INFO", "WARNING", "ERROR", "LATEST", "CVE"),
                "link": All(str, Length(min=1)),
                SchemaOptional("advisory"): All(str, Length(min=1)),
                SchemaOptional("cve_id"): All(str, Length(min=1)),
                SchemaOptional("cve_name"): All(str, Length(min=1)),
                SchemaOptional("package_name"): All(str, Length(min=1)),
                SchemaOptional("version_range"): All(str, Length(min=1)),
            }
        ]
    )

    @classmethod
    def verify_justification_schema(cls, justification: Optional[List[Dict[str, Any]]]) -> bool:
        """Verify the justification schema is correct."""
        if justification is None:
            return True

        try:
            cls._JUSTIFICATION_SCHEMA(justification)
        except Invalid as exc:
            raise AdviserJustificationSchemaError(exc.msg) from exc
        else:
            return True


class AdviserUnitTestCase(AdviserTestCase):
    """A base class for implementing pipeline unit specific test cases."""

    UNIT_TESTED: Optional[Unit] = None

    @classmethod
    def verify_multiple_should_include(cls, builder_context: PipelineBuilderContext) -> bool:
        """Check multiple should_include calls do not end in an infinite loop."""
        assert cls.UNIT_TESTED is not None, "No unit assigned for testing"
        pipeline_config = list(cls.UNIT_TESTED.should_include(builder_context))
        assert pipeline_config != [], "First call to should_include should be always non-empty generator"
        assert (
            len(pipeline_config) == 1
        ), "First call to should_include should return one config, adjust the test if it requires additional logic"

        unit = cls.UNIT_TESTED()
        unit.update_configuration(pipeline_config[0])

        builder_context.add_unit(unit)
        assert (
            list(cls.UNIT_TESTED.should_include(builder_context)) == []
        ), "Make sure the pipeline unit does not loop endlessly on multiple should_include calls"
        return True

    def test_verify_multiple_should_include(self, *args, **kwargs) -> bool:
        """Check multiple should_include calls do not end in an infinite loop."""
        # Construct a builder context that should always include a pipeline unit and pass
        # it to verify_multiple_should_include
        raise NotImplementedError(
            "Implement a test that makes sure multiple calls of should include do not loop endlessly"
        )

    def test_provided_unit_tested(self):
        """Check proper manipulation of the unit tested."""
        assert self.UNIT_TESTED is not None, "Unit tested not provided for the test base"
        assert issubclass(
            self.UNIT_TESTED, (Boot, Sieve, Pseudonym, Step, Stride, Wrap)
        ), f"Assigned unit for testing {self.UNIT_TESTED.__name__!r} does not inherit from any known pipeline unit type"

    def test_provided_package_version(self) -> None:
        """Test the unit provides package_name."""
        if self.UNIT_TESTED is None or not issubclass(self.UNIT_TESTED, (Pseudonym, Sieve, Step)):
            return pytest.skip("Unit does not type is not specific for any package version")

        unit = self.UNIT_TESTED()

        if isinstance(unit, Pseudonym):
            assert (
                isinstance(unit.configuration.get("package_name"), str) and len(unit.configuration["package_name"]) > 0
            ), (
                f"Pseudonym unit {self.UNIT_TESTED.__name__!r} does not provide required "
                "package_version configuration option"
            )
        else:
            assert (
                "package_name" in unit.configuration
            ), "It's a good practice to state package_name in the unit configuration"

            package_name = unit.configuration["package_name"]
            if package_name is not None:
                assert (
                    isinstance(package_name, str) and len(unit.configuration["package_name"]) > 0
                ), f"Unit {self.UNIT_TESTED.__name__!r} does not provide required package_version configuration option"

    def test_default_configuration(self) -> None:
        """Test instantiation the pipeline unit succeeds with default configuration."""
        unit_instance = self.UNIT_TESTED()
        unit_instance.update_configuration({})

    def test_super_pre_run(self, context: Context) -> None:
        """Make sure the pre-run method of the base is called."""
        if self.UNIT_TESTED is None or not issubclass(self.UNIT_TESTED, (Pseudonym, Sieve, Step)):
            return pytest.skip("Unit does not type is not specific for any package version")

        unit = self.UNIT_TESTED()

        assert unit.unit_run is False
        unit.unit_run = True

        with unit.assigned_context(context):
            unit.pre_run()

        assert (
            unit.unit_run is False
        ), "Unit flag unit_run not reset, is super().pre_run() called in sources when providing pre_run method!?"

    def test_default_multi_package_resolution(self) -> None:
        """Test presence of multi_package_resolution in unit default."""
        if self.UNIT_TESTED is None or not issubclass(self.UNIT_TESTED, Step):
            return pytest.skip("Only Step pipeline units should include multi_package_resolution option")

        assert (
            "multi_package_resolution" in self.UNIT_TESTED.CONFIGURATION_DEFAULT
        ), "Make sure the multi_package_resolution option is present in step unit default configuration"
        assert isinstance(self.UNIT_TESTED.CONFIGURATION_DEFAULT.get("multi_package_resolution"), bool)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages based on constraints supplied."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from packaging.specifiers import SpecifierSet
from packaging.markers import Marker
from voluptuous import Schema
from voluptuous import Required

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class ConstraintsSieve(Sieve):
    """A sieve to filter out packages based on constraints supplied."""

    CONFIGURATION_DEFAULT = {"package_name": None}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str, "specifier": str})
    _JUSTIFICATION_LINK = jl("constraints")

    packages_seen = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)
    specifier_set = attr.ib(type=SpecifierSet, default=attr.Factory(SpecifierSet), init=False)

    @classmethod
    def default_environment(cls, builder_context: "PipelineBuilderContext") -> Dict[str, str]:
        """Get environment for markers based on runtime environment configuration supplied."""
        return {
            # Not present ones based on the default environment created by packaging:
            "platform_release": "",
            "platform_version": "",
            "python_full_version": f"{builder_context.project.runtime_environment.python_version}.0",
            "implementation_name": "cpython",
            "os_name": "posix",
            "platform_machine": "x86_64",
            "platform_python_implementation": "CPython",
            "platform_system": "Linux",
            "python_version": builder_context.project.runtime_environment.python_version,
            "implementation_version": f"{builder_context.project.runtime_environment.python_version}.0",
            "sys_platform": "linux",
        }

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Enable this pipeline unit if the adjustment is enabled."""
        if builder_context.is_included(cls):
            yield from ()
            return None

        if not builder_context.project.constraints or not builder_context.project.constraints.package_versions:
            _LOGGER.debug("No constraints defined, not registering constraint sieve unit")
            yield from ()
            return None

        if builder_context.project.runtime_environment.python_version is None:
            _LOGGER.warning(
                "No Python version specified in the runtime environment, not registering constraints sieve unit"
            )
            yield from ()
            return None

        marker_default_environment = cls.default_environment(builder_context)

        for package_version in builder_context.project.constraints.package_versions.values():
            if package_version.markers:
                marker_instance = Marker(package_version.markers)
                marker_evaluation_result = marker_instance.evaluate(environment=marker_default_environment)
                if not marker_evaluation_result:
                    _LOGGER.warning(
                        "Skipping constraint on %r as marker does not apply to the environment %r",
                        package_version.name,
                        package_version.markers,
                    )
                    continue

            yield {"package_name": package_version.name, "specifier": package_version.version}

    def pre_run(self) -> None:
        """Initialize this pipeline unit before any resolution run."""
        self.packages_seen.clear()
        self.specifier_set = SpecifierSet(self.configuration.get("specifier") or "")
        # Explicitly turn on pre-releases here, as pre-releases are handled by a different pipeline unit.
        self.specifier_set.prereleases = True
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off packages that do not meet desired specifier in constraints."""
        for package_version in package_versions:
            if package_version.locked_version in self.specifier_set:
                yield package_version
                continue

            package_tuple = package_version.to_tuple()
            if package_tuple not in self.packages_seen:
                self.packages_seen.add(package_tuple)
                msg = f"Removing package {package_tuple} based on constraint {str(self.specifier_set)!r}"
                _LOGGER.info("%s - see %s", msg, self._JUSTIFICATION_LINK)
                self.context.stack_info.append(
                    {
                        "type": "INFO",
                        "message": msg,
                        "link": self._JUSTIFICATION_LINK,
                    }
                )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test changes to RHEL version for major RHEL releases."""

import flexmock

from thoth.adviser.boots import RHELVersionBoot
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.enums import RecommendationType
from thoth.python import Project

from ..base import AdviserUnitTestCase


class TestRHELVersionBoot(AdviserUnitTestCase):
    """Test changes to RHEL version for major RHEL releases."""

    _CASE_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"
"""

    UNIT_TESTED = RHELVersionBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_version_change(self) -> None:
        """Test changing RHEL version to its major version."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE))
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8.1"

        boot = RHELVersionBoot()
        with RHELVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"

    def test_version_no_change_no_os_match(self) -> None:
        """Test no change to RHEL version identifier if OS does not match."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE))
        context.project.runtime_environment.operating_system.name = "fedora"
        context.project.runtime_environment.operating_system.version = "31"

        boot = RHELVersionBoot()
        with RHELVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "fedora"
        assert context.project.runtime_environment.operating_system.version == "31"

    def test_version_no_change_no_version(self) -> None:
        """Test no change to RHEL version identifier if OS version was not supplied."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE))
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = None

        boot = RHELVersionBoot()
        with RHELVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve unit representatives used for testing.."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING
from thoth.python import PackageVersion

from thoth.adviser.sieve import Sieve
from voluptuous import Optional as SchemaOptional
from voluptuous import Required
from voluptuous import Schema

if TYPE_CHECKING:
    from thoth.adviser.pipeline_builder import PipelineBuilderContext


class Sieve1(Sieve):
    """A testing sieve implementation."""

    CONFIGURATION_DEFAULT = {"flying_circus": 1969, "package_name": "tensorflow"}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str, Required("flying_circus"): str})

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Run noop method."""
        return package_versions


class Sieve2(Sieve):
    """A testing sieve implementation."""

    CONFIGURATION_DEFAULT = {"date": "2015-09-15", "package_name": "selinon"}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("package_name"): str, Required("date"): str, SchemaOptional("foo"): str}
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Run noop method."""
        return package_versions


__all__ = ["Sieve1", "Sieve2"]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test alias pseudonym."""

import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.pseudonyms import AliasPseudonym
from thoth.adviser.context import Context
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserUnitTestCase


class TestAliasPseudonym(AdviserUnitTestCase):
    """Test related to alias pseudonym used for including a generic alias in the resolution process."""

    UNIT_TESTED = AliasPseudonym

    @pytest.mark.skip(reason="This pipeline unit is never registered implicitly")
    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""

    @pytest.mark.skip(reason="The default configuration should not be used")
    def test_default_configuration(self) -> None:
        """Check default configuration available in the pipeline unit implementation."""

    @pytest.mark.skip(reason="Package name should be always provided explicitly")
    def test_provided_package_version(self) -> None:
        """Check providing package_name in the default pipeline configuration."""

    def test_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit, never."""
        builder_context.decision_type = None
        builder_context.recommendation_type = RecommendationType.STABLE
        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    @pytest.mark.parametrize(
        "tf_version,index_url",
        [
            ("2.2.0", "https://pypi.org/simple"),
            (None, "https://pypi.org/simple"),
            ("2.2.0", None),
            (None, None),
        ],
    )
    def test_run_pseudonym(self, context: Context, tf_version: str, index_url: str) -> None:
        """Test adding a pseudonym for the given package."""
        unit = self.UNIT_TESTED()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "package_version": tf_version,
                "index_url": index_url,
                "aliases": [
                    {
                        "package_name": "intel-tensorflow",
                        "package_version": "2.2.0",
                        "index_url": "https://pypi.org/simple",
                    },
                    {
                        "package_name": "tensorflow-gpu",
                        "package_version": "2.2.0",
                        "index_url": "https://pypi.org/simple",
                    },
                    {
                        "package_name": "tensorflow",
                        "package_version": "2.2.0",
                        "index_url": "https://thoth-station.ninja/simple",
                    },
                ],
            }
        )

        package_version = PackageVersion(
            name="tensorflow", version="==2.2.0", index=Source("https://pypi.org/simple"), develop=False
        )
        result = set(unit.run(package_version))
        assert len(result) == 3
        assert result == {
            ("intel-tensorflow", "2.2.0", "https://pypi.org/simple"),
            ("tensorflow-gpu", "2.2.0", "https://pypi.org/simple"),
            ("tensorflow", "2.2.0", "https://thoth-station.ninja/simple"),
        }

    @pytest.mark.parametrize(
        "pv_version,pv_index_url,conf_version,conf_index_url",
        [
            ("==1.1.1", "https://pypi.org/simple", "2.2.0", "https://pypi.org/simple"),  # No version match.
            ("==1.1.1", "https://pypi.org/simple", "2.2.0", None),  # No version match.
            (
                "==2.2.0",
                "https://thoth-station.ninja/simple",
                "2.2.0",
                "https://pypi.org/simple",
            ),  # No index url match.
            ("==2.2.0", "https://thoth-station.ninja/simple", None, "https://pypi.org/simple"),  # No index url match.
            ("==1.1.1", "https://thoth-station.ninja/simple", "2.2.0", "https://pypi.org/simple"),  # No match.
        ],
    )
    def test_run_noop(
        self, context: Context, pv_version: str, pv_index_url: str, conf_version: str, conf_index_url: str
    ) -> None:
        """Test not adding a pseudonym for the given package."""
        unit = self.UNIT_TESTED()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "package_version": conf_version,
                "index_url": conf_index_url,
                "aliases": [
                    {
                        "package_name": "intel-tensorflow",
                        "package_version": "2.2.0",
                        "index_url": "https://pypi.org/simple",
                    },
                    {
                        "package_name": "tensorflow-gpu",
                        "package_version": "2.2.0",
                        "index_url": "https://pypi.org/simple",
                    },
                    {
                        "package_name": "tensorflow",
                        "package_version": "2.2.0",
                        "index_url": "https://thoth-station.ninja/simple",
                    },
                ],
            }
        )

        package_version = PackageVersion(
            name="tensorflow", version=pv_version, index=Source(pv_index_url), develop=False
        )
        result = list(unit.run(package_version))
        assert len(result) == 0



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test notifying about runtime environments not supported by solvers enabled in a deployment."""

from thoth.adviser.boots import SolversConfiguredBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.common import get_justification_link as jl

from ..base import AdviserUnitTestCase


class TestFullySpecifiedEnvironment(AdviserUnitTestCase):
    """Test notifying about runtime environments not supported by solvers enabled in a deployment."""

    UNIT_TESTED = SolversConfiguredBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.project.runtime_environment.operating_system.name = "rhel"
        builder_context.project.runtime_environment.operating_system.version = "8"
        builder_context.project.runtime_environment.python_version = "3.8"

        try:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = "\n\nsolver-fedora-31-py36\n\n"
            self.verify_multiple_should_include(builder_context)
        finally:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = None

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit for adviser runs."""
        builder_context.project.runtime_environment.operating_system.name = "rhel"
        builder_context.project.runtime_environment.operating_system.version = "8"
        builder_context.project.runtime_environment.python_version = "3.8"

        try:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = "\n\nsolver-fedora-31-py36\n\n"
            assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]
        finally:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = None

    def test_should_include_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test not registering the pipeline unit."""
        builder_context.project.runtime_environment.operating_system.name = "rhel"
        builder_context.project.runtime_environment.operating_system.version = "8"
        builder_context.project.runtime_environment.python_version = "3.8"

        try:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = """
                solver-fedora-31-py36\n
                solver-rhel-8-py36\n\n\n
            """
            assert list(self.UNIT_TESTED.should_include(builder_context)) == []
        finally:
            self.UNIT_TESTED._SOLVERS_CONFIGURED = None

    def test_run(self, context: Context) -> None:
        """Test if the pipeline unit properly notifies users.."""
        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            assert unit.run() is None

        assert context.stack_info == [
            {
                "link": jl("eol_env"),
                "message": "Runtime environment used is no longer supported, it is "
                "recommended to switch to another runtime environment",
                "type": "WARNING",
            },
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test boot recommending Thoth's S2I tooling."""

import pytest
from typing import Optional

from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.boots import ThothS2IBoot

from ..base import AdviserUnitTestCase


class TestThothS2IBoot(AdviserUnitTestCase):
    """Test S2I boot."""

    UNIT_TESTED = ThothS2IBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.project.runtime_environment.base_image = "rhel:8"
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize("base_image", [None, "fedora:33", "quay.io/thoth-station/solver-ubi8-py38:v0.23.0"])
    def test_include(self, base_image: Optional[str], builder_context: PipelineBuilderContext) -> None:
        """Test not including the pipeline unit."""
        builder_context.project.runtime_environment.base_image = base_image
        assert list(self.UNIT_TESTED.should_include(builder_context)) != []

    def test_run(self, context: Context) -> None:
        """Test running and recommending to use Thoth's s2i base image."""
        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with self.UNIT_TESTED.assigned_context(context):
            assert unit.run() is None

        assert len(context.stack_info) == 1
        self.verify_justification_schema(context.stack_info)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Exception hierarchy used in the whole adviser implementation."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional


class AdviserException(Exception):  # noqa: N818
    """A base adviser exception in adviser's exception hierarchy."""


class NotAcceptable(AdviserException):  # noqa: N818
    """An exception raised during stack generation when the given action would produce un-acceptable solution."""


class SkipPackage(AdviserException):  # noqa: N818
    """An exception raised during stack generation when the given package should be skipped in the stack."""


class ParseBaseImageError(AdviserException):
    """An exception raised when parsing base image fails."""


class PipfileParseError(AdviserException):
    """An exception raised on invalid Pipfile or Pipfile.lock."""


class InternalError(AdviserException):
    """An exception raised on bugs in the code base."""


class UnsupportedConfiguration(AdviserException):  # noqa: N818
    """An exception raised on unsupported configuration by recommendation engine."""


class NotFound(AdviserException):  # noqa: N818
    """An exception raised if the given resource was not found."""


class VersionIdentifierError(AdviserException):
    """An exception raised if the given version identifier is not a semver identifier."""


class UnableLock(AdviserException):  # noqa: N818
    """Raised if it is unable to lock dependencies given the set of constraints."""


class NoHistoryKept(AdviserException):  # noqa: N818
    """Raised if a user asks for history, but history was not kept (e.g. temperature function history in annealing)."""


class AdviserPipelineException(AdviserException):  # noqa: N818
    """A base class for implementing pipeline specific exceptions."""


class EagerStopPipeline(AdviserPipelineException):  # noqa: N818
    """Raised to signalize premature end of pipeline."""


class PipelineUnitError(AdviserPipelineException):
    """An exception raised when there is an error during pipeline run, unexpectedly."""


class PipelineUnitConfigurationSchemaError(PipelineUnitError):
    """An exception raised when pipeline unit configuration does not match schema declared."""


class PrescriptionSchemaError(PipelineUnitError):
    """An exception raised when prescription schema is not valid."""


class PrescriptionDuplicateUnitNameError(PipelineUnitError):
    """An exception raised when multiple prescription units share name."""


class BootError(PipelineUnitError):
    """An exception raised when pipeline boot unit fails unexpectedly."""


class SieveError(PipelineUnitError):
    """An exception raised when pipeline sieve unit fails unexpectedly."""


class StepError(PipelineUnitError):
    """An exception raised when pipeline step unit fails unexpectedly."""


class StrideError(PipelineUnitError):
    """An exception raised when pipeline stride unit fails unexpectedly."""


class WrapError(PipelineUnitError):
    """An exception raised when pipeline stride unit fails unexpectedly."""


class UnknownPipelineUnitError(PipelineUnitError):
    """An exception raised when an unknown pipeline unit is requested."""


class PipelineConfigurationError(PipelineUnitError):
    """An exception raised when a wrong pipeline unit configuration supplied.

    Or any error during configuration initialization.
    """


class AdviserRunException(AdviserException):  # noqa: N818
    """A base class for implementing exceptions occurred during an andviser run."""

    def to_dict(self) -> Optional[Dict[str, str]]:
        """Convert adviser exception to a dict representation which is shown to the user."""
        return None


class UnresolvedDependencies(AdviserRunException):  # noqa: N818
    """An exception raised if dependencies were not resolved and cannot produce stack."""

    __slots__ = ["unresolved", "stack_info"]

    def __init__(self, *args: Any, unresolved: List[str], stack_info: List[Dict[str, Any]]) -> None:
        """Capture unresolved dependencies in this exception."""
        super().__init__(*args)
        self.unresolved = unresolved
        self.stack_info = stack_info

    def to_dict(self) -> Optional[Dict[str, Any]]:
        """Convert unresolved dependencies exception to the user."""
        return {
            "ERROR": "No dependencies found for "
            f"{', '.join(f'{dep!r}' for dep in self.unresolved)}  - these "
            "dependencies were not yet solved in Thoth "
            "cannot resolve all direct dependencies",
            "_ERROR_DETAILS": {
                "unresolved": self.unresolved,
            },
            "stack_info": self.stack_info,
        }


class CannotProduceStack(AdviserRunException):  # noqa: N818
    """Raised if there was not produced any result."""

    __slots__ = ["stack_info"]

    def __init__(self, *args: Any, stack_info: List[Dict[str, Any]]) -> None:
        """Instantiate the exception."""
        super().__init__(*args)
        self.stack_info: List[Dict[str, Any]] = stack_info

    def to_dict(self) -> Optional[Dict[str, Any]]:
        """Convert exception to a dict representation for a user."""
        return {
            "ERROR": "No results were resolved, see logs for more info",
            "stack_info": self.stack_info,
        }


class UserLockFileError(AdviserRunException):
    """An exception raised when the supplied user stack has issues."""



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Filter out states randomly."""

import logging
import random
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr

from ..state import State
from ..stride import Stride
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class RandomDecisionStride(Stride):
    """Filter out states randomly."""

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Allow inclusion only per user request."""
        yield from ()
        return None

    def run(self, state: State) -> None:
        """Flip a coin and decide - tails are not acceptable."""
        if bool(random.getrandbits(1)):
            raise NotAcceptable(f"State with score {state.score!r} was randomly discarded by flipping a coin")



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of skip package sieve."""

import yaml

from flexmock import flexmock
import pytest

from thoth.adviser.exceptions import SkipPackage
from thoth.adviser.context import Context
from thoth.adviser.prescription.v1 import SkipPackageSievePrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_SKIP_PACKAGE_SIEVE_SCHEMA

from .base import AdviserUnitPrescriptionTestCase


class TestSkipPackageSievePrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to skip package sieve prescription."""

    def test_run(self, context: Context) -> None:
        """Test running this pipeline unit."""
        prescription_str = """
name: SkipPackageSieve
type: sieve.SkipPackage
should_include:
  adviser_pipeline: true
match:
  package_name: flask
run:
  stack_info:
    - type: INFO
      message: Skip package info stack info
      link: https://thoth-station.ninja
  log:
    message: Skip package info message
    type: INFO
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SKIP_PACKAGE_SIEVE_SCHEMA(prescription)
        SkipPackageSievePrescription.set_prescription(prescription)

        context.stack_info.clear()

        unit = SkipPackageSievePrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            with pytest.raises(SkipPackage):
                unit.run((i for i in (flexmock(),)))

            # Run multiple times to verify stack_info is adjusted just once.
            with pytest.raises(SkipPackage):
                unit.run((i for i in (flexmock(),)))

        assert context.stack_info == [
            {
                "type": "INFO",
                "message": "Skip package info stack info",
                "link": "https://thoth-station.ninja",
            }
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Types of units."""

from typing import Union

from .pseudonym import Pseudonym
from .boot import Boot
from .sieve import Sieve
from .step import Step
from .stride import Stride
from .wrap import Wrap
from .unit import Unit

from thoth.adviser.prescription.v1.boot import BootPrescription as BootPrescriptionV1
from thoth.adviser.prescription.v1.pseudonym import PseudonymPrescription as PseudonymPrescriptionV1
from thoth.adviser.prescription.v1.sieve import SievePrescription as SievePrescriptionV1
from thoth.adviser.prescription.v1.step import StepPrescription as StepPrescriptionV1
from thoth.adviser.prescription.v1.stride import StridePrescription as StridePrescriptionV1
from thoth.adviser.prescription.v1.unit import UnitPrescription as UnitPrescriptionV1
from thoth.adviser.prescription.v1.wrap import WrapPrescription as WrapPrescriptionV1


UnitType = Union[Unit, UnitPrescriptionV1]
BootType = Union[Boot, BootPrescriptionV1]
PseudonymType = Union[Pseudonym, PseudonymPrescriptionV1]
SieveType = Union[Sieve, SievePrescriptionV1]
StepType = Union[Step, StepPrescriptionV1]
StrideType = Union[Stride, StridePrescriptionV1]
WrapType = Union[Wrap, WrapPrescriptionV1]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap adding information about Python package releases."""

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.wraps import PyPIReleaseWrap

from ..base import AdviserUnitTestCase


class TestPyPIReleaseWrap(AdviserUnitTestCase):
    """Test adding PyPI release information."""

    UNIT_TESTED = PyPIReleaseWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run_add_justification(self) -> None:
        """Test adding link to PyPI about the given package released."""
        state = State()
        state.add_resolved_dependency(("tensorflow", "2.5.0", "https://pypi.org/simple"))

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert state.justification == [
            {
                "type": "INFO",
                "message": "Package 'tensorflow' in version '2.5.0' is released on PyPI",
                "package_name": "tensorflow",
                "link": "https://pypi.org/project/tensorflow/2.5.0/",
            }
        ]
        self.verify_justification_schema(state.justification)

    def test_run_no_justification(self) -> None:
        """Test NOT adding information if the given Python package is not a PyPI release."""
        state = State()
        state.add_resolved_dependency(("tensorflow", "2.5.0", "https://thoth-station.ninja"))

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert not state.justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of sieve prescription v1."""

import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.prescription.v1 import SievePrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_SIEVE_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestSievePrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to sieve prescription v1."""

    def test_run_stack_info(self, context: Context) -> None:
        """Check assigning stack info."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '>1.0,<=1.1.0'
    index_url: 'https://pypi.org/simple'
run:
  stack_info:
    - type: WARNING
      message: Some stack warning message printed by a sieve
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        self.check_run_stack_info(context, SievePrescription, package_versions=(pv for pv in [package_version]))

    @pytest.mark.parametrize("log_level", ["INFO", "ERROR", "WARNING"])
    def test_run_log(self, caplog, context: Context, log_level: str) -> None:
        """Check logging messages."""
        prescription_str = f"""
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '>1.0,<=1.1.0'
    index_url: 'https://pypi.org/simple'
run:
  log:
    message: Some stack warning message printed by a sieve
    type: {log_level}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        self.check_run_log(
            caplog, context, log_level, SievePrescription, package_versions=(pv for pv in [package_version])
        )

    def test_sieve(self, context: Context) -> None:
        """Test sieving packages."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '<=1.1.0'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)
        package_versions = [
            PackageVersion(
                name="flask",
                version="==1.1.0",
                index=Source("https://thoth-station.ninja"),
                develop=False,
            ),
            PackageVersion(
                name="flask",
                version="==1.1.2",
                index=Source("https://pypi.org/simple"),
                develop=False,
            ),
            PackageVersion(
                name="flask",
                version="==1.0.2",
                index=Source("https://pypi.org/simple"),
                develop=False,
            ),
        ]

        unit = SievePrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run((pv for pv in package_versions)))

        assert len(result) == 1
        assert result[0] == package_versions[1]  # ==1.1.2 stays in the result

    def test_should_include_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '<=1.1.0'
"""
        flexmock(SievePrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(SievePrescription.should_include(builder_context)) == [
            {
                "package_name": "flask",
                "match": {
                    "package_version": {
                        "name": "flask",
                        "version": "<=1.1.0",
                    },
                },
                "run": {},
            },
        ]

    def test_should_include_multi(self) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  - package_version:
      name: flask
      version: '<=1.1.0'
  - package_version:
      name: numpy
"""
        flexmock(SievePrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(SievePrescription.should_include(builder_context)) == [
            {
                "package_name": "flask",
                "match": {
                    "package_version": {
                        "name": "flask",
                        "version": "<=1.1.0",
                    },
                },
                "run": {},
            },
            {
                "package_name": "numpy",
                "match": {
                    "package_version": {
                        "name": "numpy",
                    },
                },
                "run": {},
            },
        ]

    def test_should_include_no_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    index_url: https://pypi.org/simple
"""
        flexmock(SievePrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(SievePrescription.should_include(builder_context)) == [
            {
                "package_name": None,
                "match": {
                    "package_version": {
                        "index_url": "https://pypi.org/simple",
                    }
                },
                "run": {},
            }
        ]

    def test_no_should_include(self) -> None:
        """Test not including this pipeline."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '<=1.1.0'
"""
        flexmock(SievePrescription).should_receive("_should_include_base").replace_with(lambda _: False).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(SievePrescription.should_include(builder_context)) == []

    def test_sieve_not_index_url(self, context: Context) -> None:
        """Test sieving packages based on "not" index_url."""
        prescription_str = """
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    index_url:
      not: https://pypi.org/simple
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)
        package_versions = [
            PackageVersion(
                name="flask",
                version="==1.1.0",
                index=Source("https://thoth-station.ninja/simple"),
                develop=False,
            ),
            PackageVersion(
                name="flask",
                version="==1.1.2",
                index=Source("https://pypi.org/simple"),
                develop=False,
            ),
            PackageVersion(
                name="flask",
                version="==1.0.2",
                index=Source("https://pypi.org/simple"),
                develop=False,
            ),
        ]

        unit = SievePrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run((pv for pv in package_versions)))

        assert len(result) == 2
        assert result == [package_versions[1], package_versions[2]]

    @pytest.mark.parametrize("develop", [True, False])
    def test_sieve_develop_true(self, context: Context, develop: bool) -> None:
        """Test sieving packages based on develop flag set to true."""
        prescription_str = f"""
name: SieveUnit
type: sieve
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    develop: {'true' if develop else 'false'}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SIEVE_SCHEMA(prescription)
        SievePrescription.set_prescription(prescription)
        package_versions = [
            PackageVersion(
                name="flask",
                version="==1.1.2",
                index=Source("https://pypi.org/simple"),
                develop=develop,
            ),
            PackageVersion(
                name="flask",
                version="==1.0.2",
                index=Source("https://pypi.org/simple"),
                develop=not develop,
            ),
        ]

        unit = SievePrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run((pv for pv in package_versions)))

        assert len(result) == 1
        assert [result[0]] == [pv for pv in package_versions if pv.develop != develop]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of skip package step."""

import yaml

import pytest

from thoth.adviser.exceptions import SkipPackage
from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.prescription.v1 import SkipPackageStepPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_SKIP_PACKAGE_STEP_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestSkipPackageStepPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to skip package step prescription."""

    @pytest.mark.parametrize(
        "package_version_from_version,package_version_from_index,package_version_from_develop,"
        "register_dependency,pipeline_run",
        [
            ("<=0.25.0", "https://pypi.org/simple", "false", True, True),
            (">0.25.0", "https://pypi.org/simple", "false", True, False),  # Version not matching.
            ("<=0.25.0", '{"not": "https://pypi.org/simple"}', "false", True, False),  # Index not matching.
            ("<=0.25.0", "https://pypi.org/simple", "true", True, False),  # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "false", False, False),  # Dependency is not registered.
        ],
    )
    def test_run_package_version_from(
        self,
        context: Context,
        state: State,
        package_version_from_version: str,
        package_version_from_index: str,
        package_version_from_develop: str,
        register_dependency: bool,
        pipeline_run: bool,
    ) -> None:
        """Test running the prescription based on the dependency introduced."""
        prescription_str = f"""
name: SkipPackageStep
type: step.SkipPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from:
    - name: scikit-learn
      version: "{package_version_from_version}"
      develop: {package_version_from_develop}
      index_url: {package_version_from_index}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SKIP_PACKAGE_STEP_SCHEMA(prescription)
        SkipPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        if register_dependency:
            runtime_env = context.project.runtime_environment
            context.register_package_tuple(
                package_version.to_tuple(),
                dependent_tuple=package_version_from.to_tuple(),
                develop=False,
                extras=None,
                os_name=runtime_env.operating_system.name,
                os_version=runtime_env.operating_system.version,
                python_version=runtime_env.python_version,
            )

        unit = SkipPackageStepPrescription()
        unit.pre_run()

        with unit.assigned_context(context):
            if pipeline_run:
                with pytest.raises(SkipPackage):
                    unit.run(state, package_version)
            else:
                assert unit.run(state, package_version) is None

    @pytest.mark.parametrize(
        "package_version_from_version,package_version_from_index,package_version_from_develop,"
        "resolved_version,resolved_index,resolved_develop,"
        "add_resolved,pipeline_run",
        [
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, True),
            # Version not matching.
            (">0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Index not matching.
            (
                "<=0.25.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "true", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Version not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "!=8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Index not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "true", True, False),
            # Resolved dependency nit present.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                False,
                False,
            ),
        ],
    )
    def test_run_package_version_from_with_resolved(
        self,
        context: Context,
        state: State,
        package_version_from_version: str,
        package_version_from_index: str,
        package_version_from_develop: str,
        resolved_version: str,
        resolved_index: str,
        resolved_develop: str,
        add_resolved: bool,
        pipeline_run: bool,
    ) -> None:
        """Test running the prescription based on the dependency introduced."""
        prescription_str = f"""
name: SkipPackageStep
type: step.SkipPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from:
    - name: scikit-learn
      version: "{package_version_from_version}"
      develop: {package_version_from_develop}
      index_url: {package_version_from_index}
    resolved_dependencies:
    - name: click
      version: "{resolved_version}"
      develop: {resolved_develop}
      index_url: {resolved_index}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_SKIP_PACKAGE_STEP_SCHEMA(prescription)
        SkipPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        if add_resolved:
            state.add_resolved_dependency(package_version_resolved.to_tuple())
            context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        unit = SkipPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            if pipeline_run:
                with pytest.raises(SkipPackage):
                    unit.run(state, package_version)
            else:
                assert unit.run(state, package_version) is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A pseudonym that introduces package aliases based on the supplied configuration."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Required
from voluptuous import Schema

from ...pseudonym import Pseudonym

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class AliasPseudonym(Pseudonym):
    """A pseudonym that introduces package aliases based on the supplied configuration."""

    CONFIGURATION_DEFAULT = {"package_name": None, "package_version": None, "index_url": None}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): str,
            Required("package_version"): SchemaAny(str, None),
            Required("index_url"): SchemaAny(str, None),
            Required("aliases"): Schema(
                [Schema({Required("package_name"): str, Required("package_version"): str, Required("index_url"): str})]
            ),
        }
    )

    @classmethod
    def should_include(cls, _: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, never."""
        yield from ()
        return None

    def run(self, package_version: PackageVersion) -> Generator[Tuple[str, str, str], None, None]:
        """Create alternatives to packages based on the configuration supplied."""
        if (
            self.configuration["package_version"] is not None
            and self.configuration["package_version"] != package_version.locked_version
        ):
            yield from ()
            return

        if self.configuration["index_url"] is not None and self.configuration["index_url"] != package_version.index.url:
            yield from ()
            return

        yield from ((i["package_name"], i["package_version"], i["index_url"]) for i in self.configuration["aliases"])



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out legacy versions."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class LegacyVersionSieve(Sieve):
    """A sieve to filter out legacy versions.

    This sieve assumes no recent projects use legacy versions hence we can
    simply skip them in the resolution process.
    """

    CONFIGURATION_DEFAULT = {"package_name": None}

    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Include this sieve once, always."""
        if not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._messages_logged.clear()
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off legacy versions from the resolution process."""
        for package_version in package_versions:
            if package_version.semantic_version.is_legacy_version:
                package_tuple = package_version.to_tuple()
                if package_tuple not in self._messages_logged:
                    self._messages_logged.add(package_tuple)
                    _LOGGER.warning(
                        "Removing package %s as the version identifier is a legacy version string",
                        package_version.to_tuple(),
                    )
                continue

            yield package_version
